# Default variables
DEFAULT_RPC_URL="http://18.188.176.119:8545"
DEFAULT_OG_FAUCET_URL="https://faucet.opengradient.ai/?address="
DEFAULT_HUB_SIGNUP_URL="https://hub.opengradient.ai/signup"
DEFAULT_INFERENCE_CONTRACT_ADDRESS="0x8383C9bD7462F12Eb996DD02F78234C0421A6FaE"
DEFAULT_BLOCKCHAIN_EXPLORER="https://explorer.opengradient.ai/tx/"
DEFAULT_IMAGE_GEN_HOST="18.217.25.69"
DEFAULT_IMAGE_GEN_PORT=5125
